import 'dart:mirrors';

String symbolToName(Symbol symbol) {
  final bs = symbol.toString();
  return bs.substring(8, bs.length - 2);
}

String nameUnderLine(String original) {
  return original
      .split('')
      .map((w) {
        if (w.toLowerCase() == w) {
          return w;
        } else {
          return '_${w.toLowerCase()}';
        }
      })
      .toList()
      .join('');
}

class Mapper<M> {
  Mapper();

  M? object(Map<String, dynamic> map) {
    return mapToObject(M, map);
  }

  dynamic mapToObject(Type cls, Map<String, dynamic> map) {
    final cm = reflectClass(cls);
    final params = <Symbol, dynamic>{};
    cm.instanceMembers.forEach((Symbol key, MethodMirror value) {
      if (value.isSynthetic && value.isGetter) {
        final mapKey = nameUnderLine(symbolToName(key));
        final fieldValue = map[mapKey];
        if ([int, double, bool, String]
            .contains(value.returnType.reflectedType)) {
          if (map.containsKey(mapKey)) {
            if (fieldValue != null) {
              params[key] = fieldValue;
            }
          }
        } else {
          if (map.containsKey(mapKey)) {
            if (fieldValue != null) {
              final objMap =
                  mapToObject(value.returnType.reflectedType, fieldValue);
              params[key] = objMap;
            }
          }
        }
      }
    });
    final ins = cm.newInstance(const Symbol(''), [], params);
    return ins.reflectee;
  }

  Map<String, dynamic> toMap(M object) {
    return objectToMap(M, object);
  }

  Map<String, dynamic> objectToMap(Type cls, dynamic object) {
    final om = reflect(object);
    final cm = reflectClass(cls);
    final map = <String, dynamic>{};
    cm.instanceMembers.forEach((Symbol key, MethodMirror value) {
      if (value.isSynthetic && value.isGetter) {
        final mapKey = nameUnderLine(symbolToName(key));
        if ([int, double, bool, String]
            .contains(value.returnType.reflectedType)) {
          map[mapKey] = om.getField(key).reflectee;
        } else {
          final subObj = om.getField(key).reflectee;
          final subMap = objectToMap(value.returnType.reflectedType, subObj);
          map[mapKey] = subMap;
        }
      }
    });
    return map;
  }
}
